/**
 * @author [Tristan Valcke]{@link https://github.com/Itee}
 * @license [BSD-3-Clause]{@link https://opensource.org/licenses/BSD-3-Clause}
 *
 * @module Modules/UploadDispatcher
 *
 * @description This module will dispatch the uplaod file to the correct parser.
 */

const I     = require( 'i-return' )
const path  = require( 'path' )
const THREE = require( 'three-full' )

// Todo: Make it a converter
// const AscFile = require( './AscFile' )

const JsonToThree = require( './JsonToThree' )
const ShpToThree  = require( './ShpToThree' )
const DbfToThree  = require( './DbfToThree' )
const MtlToThree  = require( './MtlToThree' )
const ObjToThree  = require( './ObjToThree' )
const Obj2ToThree = require( './Obj2ToThree' )

const ThreeToMongoDB = require( './ThreeToMongoDB' )

const FileFormat = Object.freeze( {
    Asc:  '.asc',
    Dbf:  '.dbf',
    Fbx:  '.fbx',
    Mtl:  '.mtl',
    Json: '.json',
    Obj:  '.obj',
    Shp:  '.shp',
    Stl:  '.stl',

    toString () {

        const formats = Object.values( this )
        let result    = ''
        for ( let index = 0, numberOfFormats = formats.length ; index < numberOfFormats ; index++ ) {
            result += formats[ index ]
            result += ( ( index === numberOfFormats - 1 ) ? ', ' : '.' )
        }

    }
} )

// Todo: allow to add/remove converters and output database
class FilesToDatabase {

    constructor ( Mongoose ) {

        this._numberOfFileToProcess = 0
        this._errors                = []

        this._jsonToThree = new JsonToThree()
        this._shpToThree  = new ShpToThree()
        this._dbfToThree  = new DbfToThree()
        this._mtlToThree  = new MtlToThree()
        this._objToThree  = ( true ) ? new ObjToThree() : new Obj2ToThree()

        this._parentId        = undefined
        this._disableRootNode = undefined
        this._threeToMongo    = new ThreeToMongoDB( Mongoose )

    }

    // Public
    saveFilesInDatabase ( request, response ) {

        const parameters    = request.body
        const files         = this._convertFilesObjectToArray( request.files )
        const numberOfFiles = files.length

        this._parentId        = parameters.parentId
        this._disableRootNode = ( parameters.disableRootNode === 'true' )

        if ( numberOfFiles === 0 ) {

            I.returnData( null, response )

        } else if ( numberOfFiles === 1 ) {

            this._saveSingleFile( files[ 0 ], parameters, response )

        } else if ( numberOfFiles === 2 ) {

            this._saveAssociativeFiles( files[ 0 ], files[ 1 ], parameters, response )

        } else {

            I.returnError( `Impossible d'analyser ${numberOfFiles} fichiers associatifs simultanément !`, response )

        }

    }

    // Private
    _convertFilesObjectToArray ( files ) {

        const fileArray = []

        for ( let field in files ) {

            if ( files.hasOwnProperty( field ) ) {

                fileArray.push( files[ field ] )

            }

        }

        return fileArray

    }

    _fileConversionSuccessCallback ( response, extraSuccessCallback, data ) {

        if ( extraSuccessCallback ) {
            extraSuccessCallback( data )
            return
        }

        this._threeToMongo.save(
            data,
            {
                parentId:        this._parentId,
                disableRootNode: this._disableRootNode
            },
            ( success ) => {
                this._numberOfFileToProcess--
                this._checkEndOfReturns( response, [
                    {
                        title:   'Succées',
                        message: `Sauvegarder sous l'identifiant ${success}`
                    }
                ] )
            },
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _fileConversionProgressCallback ( response, progress ) {

        console.log( progress )

    }

    _fileConversionErrorCallback ( response, error ) {

        this._errors.push( error )
        this._numberOfFileToProcess--
        this._checkEndOfReturns( response )

    }

    _checkEndOfReturns ( response, data ) {

        if ( this._numberOfFileToProcess > 0 ) { return }

        if ( this._errors.length > 0 ) {
            I.return( response )( this._errors )
            this._errors = []
        } else {
            I.returnData( data, response )
        }

    }

    ///

    _saveSingleFile ( file, parameters, response ) {

        const fileExtension = path.extname( file.filename )

        this._numberOfFileToProcess++

        switch ( fileExtension ) {

            case FileFormat.Asc:
                this._saveAsc( file.file, parameters, response )
                break

            case FileFormat.Obj:
                this._saveObj( file.file, parameters, response )
                break

            case FileFormat.Mtl:
                this._saveMtl( file.file, parameters, response )
                break

            case FileFormat.Json:
                this._saveJson( file.file, parameters, response )
                break

            case FileFormat.Shp:
                this._saveShp( file.file, parameters, response )
                break

            default:
                this._errors.push( {
                    title:   'Mauvaise extension de fichier',
                    message: `Le format de fichier ${fileExtension} n'est pas géré !`
                } )
                this._numberOfFileToProcess--
                this._checkEndOfReturns( response )
                break

        }

    }

    _saveAsc ( file, parameters, response ) {

        AscFile.parse( file, parameters, this._fileConversionErrorCallback.bind( this, response ) )

    }

    _saveObj ( file, parameters, response ) {

        this._objToThree.convert(
            file,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, null ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _saveMtl ( file, parameters, response ) {

        this._mtlToThree.convert(
            file,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, null ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _saveJson ( file, parameters, response ) {

        this._jsonToThree.convert(
            file,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, null ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _saveShp ( file, parameters, response ) {

        this._shpToThree.convert(
            file,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, null ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    ///

    _saveAssociativeFiles ( firstFile, secondFile, parameters, response ) {

        // Check files extensions
        const firstFileExtension  = path.extname( firstFile.filename )
        const secondFileExtension = path.extname( secondFile.filename )

        this._numberOfFileToProcess++ // Unecessary to set to 2 due to the fact that files are considered as a single package

        if ( firstFileExtension === FileFormat.Mtl && secondFileExtension === FileFormat.Obj ) {

            this._saveObjMtlCouple( secondFile.file, firstFile.file, parameters, response )

        } else if ( firstFileExtension === FileFormat.Obj && secondFileExtension === FileFormat.Mtl ) {

            this._saveObjMtlCouple( firstFile.file, secondFile.file, parameters, response )

        } else if ( firstFileExtension === FileFormat.Shp && secondFileExtension === FileFormat.Dbf ) {

            this._saveShpDbfCouple( firstFile.file, secondFile.file, parameters, response )

        } else if ( firstFileExtension === FileFormat.Dbf && secondFileExtension === FileFormat.Shp ) {

            this._saveShpDbfCouple( secondFile.file, firstFile.file, parameters, response )

        } else {

            this._errors.push( {
                title:   'Mauvaise extension de fichier',
                message: `Extension de fichiers associaifs inconnue: ${firstFileExtension}, ${secondFileExtension}`
            } )
            this._numberOfFileToProcess--
            this._checkEndOfReturns( response )

        }

    }

    _saveObjMtlCouple ( objFile, mtlFile, parameters, response ) {

        this._mtlToThree.convert(
            mtlFile,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, function ( materials ) {

                materials.preload()
                parameters[ 'materials' ] = materials

                this._objToThree.convert(
                    objFile,
                    parameters,
                    this._fileConversionSuccessCallback.bind( this, response, null ),
                    this._fileConversionProgressCallback.bind( this, response ),
                    this._fileConversionErrorCallback.bind( this, response )
                )

            }.bind( this ) ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _saveShpDbfCouple ( shpFile, dbfFile, parameters, response ) {

        const self = this

        this._shpToThree.convert(
            shpFile,
            parameters,
            this._fileConversionSuccessCallback.bind( this, response, function ( shpData ) {

                this._dbfToThree.convert( dbfFile,
                    parameters,
                    dbfData => {

                        const group = new THREE.Group()
                        group.name  = 'Locaux'

                        let mesh     = undefined
                        let geometry = undefined
                        let material = undefined
                        for ( let shapeIndex = 0, numberOfShapes = shpData.length ; shapeIndex < numberOfShapes ; shapeIndex++ ) {

                            geometry = new THREE.ShapeBufferGeometry( shpData[ shapeIndex ] )
                            material = new THREE.MeshPhongMaterial( {
                                color: 0xb0f2b6,
                                side:  THREE.DoubleSide
                            } )
                            mesh     = new THREE.Mesh( geometry, material )

                            // Merge Dbf data with Shp data
                            const shapeName         = dbfData.datas[ shapeIndex ][ 'CODE' ]
                            mesh.name               = shapeName
                            mesh.userData[ 'Code' ] = shapeName

                            group.add( mesh )

                        }

                        // Insert to db here !
                        self._fileConversionSuccessCallback( response, null, group )

                    },
                    this._fileConversionProgressCallback.bind( this, response ),
                    this._fileConversionErrorCallback.bind( this, response )
                )

            }.bind( this ) ),
            this._fileConversionProgressCallback.bind( this, response ),
            this._fileConversionErrorCallback.bind( this, response )
        )

    }

    _saveThreeInMongoDB ( threeData ) {

        console.log( threeData )

    }

}

module.exports = FilesToDatabase
