/**
 * @author [Tristan Valcke]{@link https://github.com/Itee}
 * @license [BSD-3-Clause]{@link https://opensource.org/licenses/BSD-3-Clause}
 *
 * @class Todo...
 * @classdesc Todo...
 * @example Todo...
 *
 */

/* eslint-env browser */

const { Enum }                                               = require( 'enumify' )
const { isNull, isUndefined, isNotNumber, isNumberNegative } = require( 'itee-validators' )

/**
 *
 * @type {Object}
 */
class Endianness extends Enum {}

Endianness.initEnum( {
    Little: true,
    Big:    false
} )

/**
 *
 * @type {Object}
 */
class Byte extends Enum {}

Byte.initEnum( {
    One:    1,
    Two:    2,
    Four:   4,
    Height: 8
} )

class TBinaryReader {

    constructor ( buffer = new ArrayBuffer( 0 ), offset = 0, length = 0, endianness = Endianness.Little ) {

        this.buffer     = buffer
        this.offset     = offset
        this.length     = length
        this.endianness = endianness

        this._updateDataView()

    }

    get buffer () {

        return this._buffer

    }

    set buffer ( value ) {

        if ( isNull( value ) ) { throw new TypeError( 'Buffer cannot be null ! Expect an instance of ArrayBuffer.' ) }
        if ( isUndefined( value ) ) { throw new TypeError( 'Buffer cannot be undefined ! Expect an instance of ArrayBuffer.' ) }
        if ( !( value instanceof ArrayBuffer ) ) { throw new TypeError( `Buffer cannot be an instance of ${value.constructor.name} ! Expect an instance of ArrayBuffer.` ) }

        this._buffer = value
        this._offset = 0
        this._length = this._buffer.length

        this._updateDataView()

    }

    get offset () {

        return this._offset

    }

    set offset ( value ) {

        if ( isNull( value ) ) { throw new TypeError( 'Offset cannot be null ! Expect zero or a positive number.' ) }
        if ( isUndefined( value ) ) { throw new TypeError( 'Offset cannot be undefined ! Expect zero or a positive number.' ) }
        if ( isNotNumber( value ) ) { throw new TypeError( `Offset cannot be an instance of ${value.constructor.name} ! Expect zero or a positive number.` ) }
        if ( isNumberNegative( value ) ) { throw new TypeError( 'Offset cannot be lesser than zero ! Expect zero or a positive number.' ) }

        this._offset = value

        this._updateDataView()

    }

    get length () {

        return this._length

    }

    set length ( value ) {

        if ( isNull( value ) ) { throw new TypeError( 'Length cannot be null ! Expect zero or a positive number.' ) }
        if ( isUndefined( value ) ) { throw new TypeError( 'Length cannot be undefined ! Expect zero or a positive number.' ) }
        if ( isNotNumber( value ) ) { throw new TypeError( `Length cannot be an instance of ${value.constructor.name} ! Expect zero or a positive number.` ) }
        if ( isNumberNegative( value ) ) { throw new TypeError( 'Length cannot be lesser than zero ! Expect zero or a positive number.' ) }

        this._length = value

        this._updateDataView()

    }

    get endianness () {

        return this._endianness

    }

    set endianness ( value ) {

        if ( isNull( value ) ) { throw new TypeError( 'Endianness cannot be null ! Expect an enum value of Endianness.' ) }
        if ( isUndefined( value ) ) { throw new TypeError( 'Endianness cannot be undefined ! Expect an enum value of Endianness.' ) }
        if ( !( value instanceof Endianness ) ) { throw new TypeError( `Buffer cannot be an instance of ${value.constructor.name} ! Expect an enum value of Endianness.` ) }

        this._endianness = value

    }

    setBuffer ( value ) {

        this.buffer = value
        return this

    }

    setOffset ( value ) {

        this.offset = value
        return this

    }

    setLength ( value ) {

        this.length = value
        return this

    }

    setEndianness ( value ) {

        this.endianness = value
        return this

    }

    //// Methods

    /**
     *
     * @param increment
     * @return {*}
     * @private
     */
    _getAndUpdateOffsetBy ( increment ) {

        const currentOffset = this._offset
        this._offset += increment
        return currentOffset

    }

    /**
     *
     * @private
     */
    _updateDataView () {

        this._dataView = new DataView( this._buffer, this._offset, this._length )

    }

    /**
     *
     * @return {boolean}
     */
    isEndOfFile () {

        return ( this._offset === this._length )

    }

    /**
     *
     * @param offset
     * @return {skipOffsetTo}
     */
    skipOffsetTo ( offset ) {

        this._offset = offset

        return this

    }

    /**
     *
     * @param nBytes
     * @return {skipOffsetOf}
     */
    skipOffsetOf ( nBytes ) {

        this._offset += nBytes

        return this

    }

    /**
     *
     * @param buffer
     * @param offset
     * @param length
     * @return {setBuffer}
     */
    setBuffer ( buffer, offset = 0, length = buffer.byteLength ) {

        this._buffer = buffer
        this._offset = offset
        this._length = length

        this._updateDataView()

        return this

    }

    /**
     *
     * @return {number}
     */
    getInt8 () {

        return this._dataView.getInt8( this._getAndUpdateOffsetBy( Byte.One ) )

    }

    /**
     *
     * @return {number}
     */
    getUInt8 () {

        return this._dataView.getUint8( this._getAndUpdateOffsetBy( Byte.One ) )

    }

    /**
     *
     * @return {number}
     */
    getInt16 () {

        return this._dataView.getInt16( this._getAndUpdateOffsetBy( Byte.Two ), this._endianness )

    }

    /**
     *
     * @return {number}
     */
    getUInt16 () {

        return this._dataView.getUint16( this._getAndUpdateOffsetBy( Byte.Two ), this._endianness )

    }

    /**
     *
     * @return {number}
     */
    getInt32 () {

        return this._dataView.getInt32( this._getAndUpdateOffsetBy( Byte.Four ), this._endianness )

    }

    /**
     *
     * @return {number}
     */
    getUInt32 () {

        return this._dataView.getUint32( this._getAndUpdateOffsetBy( Byte.Four ), this._endianness )

    }

    /**
     *
     * @return {number}
     */
    getFloat () {

        return this._dataView.getFloat32( this._getAndUpdateOffsetBy( Byte.Four ), this._endianness )

    }

    /**
     *
     * @return {number}
     */
    getDouble () {

        return this._dataView.getFloat64( this._getAndUpdateOffsetBy( Byte.Height ), this._endianness )

    }

    /**
     *
     * @return {string}
     */
    getChar () {

        return String.fromCharCode( this._dataView.getInt8( this._getAndUpdateOffsetBy( Byte.One ) ) )

    }

    /**
     *
     * @param length
     * @param trim
     * @return {string}
     */
    getString ( length, trim = true ) {

        let string   = ''
        let charCode = undefined

        for ( let i = 0 ; i < length ; i++ ) {
            charCode = this.getUInt8()

            if ( charCode === 0 ) {
                continue
            }

            string += String.fromCharCode( charCode )
        }

        if ( trim ) {
            string = string.trim()
        }

        return string

    }

}

module.exports = TBinaryReader
